from knowledge_mapper.knowledge_base import KnowledgeBase

class DataSource:
    def test(self):
        raise NotImplementedError("Please implement this abstract method.")
    def handle(self, ki, binding_set, requesting_kb):
        raise NotImplementedError("Please implement this abstract method.")
    
    def set_knowledge_base(self, kb: KnowledgeBase):
        self.kb = kb



from __future__ import annotations
from dataclasses import dataclass, field
from typing import Callable

import requests

from knowledge_mapper.tke_exceptions import UnexpectedHttpResponseError

ASK = 'AskKnowledgeInteraction'
ANSWER = 'AnswerKnowledgeInteraction'
POST = 'PostKnowledgeInteraction'
REACT = 'ReactKnowledgeInteraction'

@dataclass(kw_only=True)
class KnowledgeInteractionRegistrationRequest:
    prefixes: dict = field(default_factory=dict)
    type: str = None # This is defined in the concrete child classes.

@dataclass(kw_only=True)
class AskKnowledgeInteractionRegistrationRequest(KnowledgeInteractionRegistrationRequest):
    pattern: str
    type: str = ASK

@dataclass(kw_only=True)
class AnswerKnowledgeInteractionRegistrationRequest(KnowledgeInteractionRegistrationRequest):
    pattern: str
    type: str = ANSWER
    handler: Callable[[list[dict], str], list[dict]]

@dataclass(kw_only=True)
class PostKnowledgeInteractionRegistrationRequest(KnowledgeInteractionRegistrationRequest):
    argument_pattern: str
    result_pattern: str
    type: str = POST

@dataclass(kw_only=True)
class ReactKnowledgeInteractionRegistrationRequest(KnowledgeInteractionRegistrationRequest):
    argument_pattern: str
    result_pattern: str
    type: str = REACT
    handler: Callable[[list[dict], str], list[dict]]

class KnowledgeInteraction:
    def __init__(self, id: str, type: str, kb, name=None):
        self.id = id
        self.type = type
        self.kb = kb
        self.name = name

    def from_req(req: KnowledgeInteractionRegistrationRequest, id: str, kb, name: str=None) -> KnowledgeInteraction:
        if isinstance(req, AskKnowledgeInteractionRegistrationRequest):
            return AskKnowledgeInteraction(req, id, kb, name)
        elif isinstance(req, AnswerKnowledgeInteractionRegistrationRequest):
            return AnswerKnowledgeInteraction(req, id, kb, name)
        elif isinstance(req, PostKnowledgeInteractionRegistrationRequest):
            return PostKnowledgeInteraction(req, id, kb, name)
        elif isinstance(req, ReactKnowledgeInteractionRegistrationRequest):
            return ReactKnowledgeInteraction(req, id, kb, name)
        else:
            raise Exception('`req` must be a concrete knowledge interaction object')
    
    def __eq__(self, other: object) -> bool:
        return self.id == other.id and self.type == other.type and self.name == other.name


class AskKnowledgeInteraction(KnowledgeInteraction):
    def __init__(self, req: AskKnowledgeInteractionRegistrationRequest, id: str, kb, name=None):
        super().__init__(id, req.type, kb)
        self.pattern = req.pattern
        self.prefixes = req.prefixes

    def ask(self, bindings: dict) -> dict:
        response = requests.post(
            f'{self.kb.ke_url}/sc/ask',
            json=bindings,
            headers={
                'Knowledge-Base-Id': self.kb.id,
                'Knowledge-Interaction-Id': self.id,
            }
        )
        if not response.ok:
            raise UnexpectedHttpResponseError(response)

        return response.json()


class AnswerKnowledgeInteraction(KnowledgeInteraction):
    def __init__(self, req: AnswerKnowledgeInteractionRegistrationRequest, id: str, kb, name=None):
        super().__init__(id, req.type, kb)
        self.pattern = req.pattern
        self.prefixes = req.prefixes
        self.handler = req.handler

    def answer(self, bindings: list[dict], requesting_kb_id: str) -> list[dict]:
        return self.handler(bindings, requesting_kb_id)


class PostKnowledgeInteraction(KnowledgeInteraction):
    def __init__(self, req: PostKnowledgeInteractionRegistrationRequest, id: str, kb, name=None):
        super().__init__(id, req.type, kb)
        self.argument_pattern = req.argument_pattern
        self.result_pattern = req.result_pattern
        self.prefixes = req.prefixes

    def post(self, bindings: list[dict]) -> list[dict]:
        response = requests.post(
            f'{self.kb.ke_url}/sc/post',
            json=bindings,
            headers={
                'Knowledge-Base-Id': self.kb.id,
                'Knowledge-Interaction-Id': self.id,
            }
        )
        if not response.ok:
            raise UnexpectedHttpResponseError(response)

        return response.json()


class ReactKnowledgeInteraction(KnowledgeInteraction):
    def __init__(self, req: ReactKnowledgeInteractionRegistrationRequest, id: str, kb, name=None):
        super().__init__(id, req.type, kb)
        self.argument_pattern = req.argument_pattern
        self.result_pattern = req.result_pattern
        self.prefixes = req.prefixes
        self.handler = req.handler

    def react(self, bindings: list[dict], requesting_kb_id: str) -> list[dict]:
        return self.handler(bindings, requesting_kb_id)



import requests
import logging as log
import time

import knowledge_mapper.knowledge_base as knowledge_base

from knowledge_mapper.tke_exceptions import UnexpectedHttpResponseError

MAX_CONNECTION_ATTEMPTS = 10
WAIT_BEFORE_RETRY = 1

class TkeClient:
    def __init__(self, ke_url: str):
        self.ke_url = ke_url


    def connect(self, max_attempts=MAX_CONNECTION_ATTEMPTS, wait_between_attempts=WAIT_BEFORE_RETRY):
        attempts = 0
        success = False
        while not success:
            try:
                attempts += 1
                self.get_knowledge_bases()
                success = True
            except requests.exceptions.ConnectionError:
                log.warning(f'Connecting to {self.ke_url} failed.')

            if not success and (max_attempts is None or attempts < max_attempts):
                log.warning(f'Request to {self.ke_url} failed after attempt {attempts}. Retrying in {wait_between_attempts} s.')
                time.sleep(wait_between_attempts)
            elif not success:
                raise Exception(f'Request to {self.ke_url} failed. Gave up after {attempts} attempts.')
        log.info(f'Successfully connected to {self.ke_url}.')


    def get_knowledge_bases(self) -> list[knowledge_base.KnowledgeBase]:
        response = requests.get(f'{self.ke_url}/sc')

        if not response.ok:
            raise UnexpectedHttpResponseError(response)

        return [knowledge_base.KnowledgeBase.from_json(kb_data, self.ke_url) for kb_data in response.json()]

    def get_knowledge_base(self, kb_id: str) -> knowledge_base.KnowledgeBase | None:
        response = requests.get(
            f'{self.ke_url}/sc',
            headers={
                'Knowledge-Base-Id': kb_id
            }
        )

        if response.status_code == 404:
            return None
        elif not response.ok:
            raise UnexpectedHttpResponseError(response)

        return knowledge_base.KnowledgeBase.from_json(response.json()[0], self.ke_url)


    def register(self, req: knowledge_base.KnowledgeBaseRegistrationRequest, reregister=True) -> knowledge_base.KnowledgeBase | None:
        already_existing = self.get_knowledge_base(req.id)
        if already_existing is not None:
            if reregister:
                already_existing.unregister()
            else:
                return None

        response = requests.post(
            f'{self.ke_url}/sc',
            json={
                'knowledgeBaseId': req.id,
                'knowledgeBaseName': req.name,
                'knowledgeBaseDescription': req.description
            }
        )
        if not response.ok:
            raise UnexpectedHttpResponseError(response)

        return knowledge_base.KnowledgeBase(req, self.ke_url)


class CleanUpFailedError(RuntimeError):
    pass



from __future__ import annotations
from dataclasses import dataclass
import logging as log
import requests

import knowledge_mapper.knowledge_interaction as knowledge_interaction

from knowledge_mapper.tke_exceptions import UnexpectedHttpResponseError


class KnowledgeBaseUnregistered(Exception):
    pass


@dataclass(kw_only=True)
class KnowledgeBaseRegistrationRequest:
    """Class with necessary data itemsfor registering a new knowledge base."""

    id: str
    name: str
    description: str


class KnowledgeBase:
    def __init__(self, req: KnowledgeBaseRegistrationRequest, ke_url: str):
        self.ke_url = ke_url
        self.id = req.id
        self.name = req.name
        self.description = req.description
        self.kis = dict()
        self.kis_by_name = dict()

    def from_json(kb_json: dict, ke_url: str) -> KnowledgeBase:
        kb = KnowledgeBase(
            KnowledgeBaseRegistrationRequest(
                id=kb_json["knowledgeBaseId"],
                name=kb_json["knowledgeBaseName"],
                description=kb_json["knowledgeBaseDescription"],
            ),
            ke_url=ke_url,
        )

        kb.sync_knowledge_interactions()

        return kb

    def unregister(self):
        if self.ke_url is None:
            raise Exception(
                "Cannot unregister this KB because no knowledge engine URL is known for this object."
            )

        response = requests.delete(
            f"{self.ke_url}/sc", headers={"Knowledge-Base-Id": self.id}
        )

        if not response.ok:
            raise UnexpectedHttpResponseError(response)

    def register_knowledge_interaction(
        self,
        ki: knowledge_interaction.KnowledgeInteractionRegistrationRequest,
        name=None,
    ) -> knowledge_interaction.KnowledgeInteraction:
        body = {"knowledgeInteractionType": ki.type, "prefixes": ki.prefixes}

        if name is not None:
            body["knowledgeInteractionName"] = name

        if isinstance(
            ki,
            knowledge_interaction.AskKnowledgeInteractionRegistrationRequest
            | knowledge_interaction.AnswerKnowledgeInteractionRegistrationRequest,
        ):
            body["graphPattern"] = ki.pattern
        elif isinstance(
            ki,
            knowledge_interaction.PostKnowledgeInteractionRegistrationRequest
            | knowledge_interaction.ReactKnowledgeInteractionRegistrationRequest,
        ):
            body["argumentGraphPattern"] = ki.argument_pattern
            body["resultGraphPattern"] = ki.result_pattern
        else:
            raise Exception("`ki` must be a concrete knowledge interaction object")

        response = requests.post(
            f"{self.ke_url}/sc/ki", headers={"Knowledge-Base-Id": self.id}, json=body
        )

        if not response.ok:
            raise UnexpectedHttpResponseError(response)

        ki_id = response.json()["knowledgeInteractionId"]

        log.info(f"Successfully registered knowledge interaction {ki_id}.")

        registered_ki = knowledge_interaction.KnowledgeInteraction.from_req(
            ki, ki_id, self
        )
        self.kis[ki_id] = registered_ki

        if name is not None:
            self.kis_by_name[name] = registered_ki

        return registered_ki

    def sync_knowledge_interactions(self):
        response = requests.get(
            f"{self.ke_url}/sc/ki", headers={"Knowledge-Base-Id": self.id}
        )

        if not response.ok:
            raise UnexpectedHttpResponseError(response)

        raw_kis = response.json()
        kis_not_seen_yet = self.kis.copy()
        kis_by_name_not_seen_yet = self.kis_by_name.copy()
        for raw_ki in raw_kis:
            if raw_ki["knowledgeInteractionType"] == "AskKnowledgeInteraction":
                req = knowledge_interaction.AskKnowledgeInteractionRegistrationRequest(
                    pattern=raw_ki["graphPattern"]
                )
            elif raw_ki["knowledgeInteractionType"] == "AnswerKnowledgeInteraction":
                req = (
                    knowledge_interaction.AnswerKnowledgeInteractionRegistrationRequest(
                        pattern=raw_ki["graphPattern"], handler=None
                    )
                )
            elif raw_ki["knowledgeInteractionType"] == "PostKnowledgeInteraction":
                req = knowledge_interaction.PostKnowledgeInteractionRegistrationRequest(
                    argument_pattern=raw_ki["argumentGraphPattern"],
                    result_pattern=raw_ki["resultGraphPattern"]
                    if "resultGraphPattern" in raw_ki
                    else None,
                )
            elif raw_ki["knowledgeInteractionType"] == "ReactKnowledgeInteraction":
                req = (
                    knowledge_interaction.ReactKnowledgeInteractionRegistrationRequest(
                        argument_pattern=raw_ki["argumentGraphPattern"],
                        result_pattern=raw_ki["resultGraphPattern"]
                        if "resultGraphPattern" in raw_ki
                        else None,
                        handler=None,
                    )
                )
            else:
                raise Exception(
                    f"Invalid KI type: {raw_ki['knowledgeInteractionType']}"
                )

            if "knowledgeInteractionName" in raw_ki:
                name = raw_ki["knowledgeInteractionName"]
            else:
                name = None

            ki_id = raw_ki["knowledgeInteractionId"]

            ki = knowledge_interaction.KnowledgeInteraction.from_req(
                req, ki_id, self, name
            )

            self.kis[ki_id] = ki
            if ki_id in kis_not_seen_yet:
                del kis_not_seen_yet[ki_id]
            if name is not None:
                self.kis_by_name[name] = ki
                if name in kis_by_name_not_seen_yet:
                    del kis_by_name_not_seen_yet[name]

        for no_longer_existing_ki_id in kis_not_seen_yet.keys():
            del self.kis[no_longer_existing_ki_id]

        for no_longer_existing_ki_name in kis_by_name_not_seen_yet.keys():
            del self.kis_by_name[no_longer_existing_ki_name]

    def get_ki(self, name=None, id=None) -> knowledge_interaction.KnowledgeInteraction:
        if name is not None and id is None:
            if name in self.kis_by_name:
                return self.kis_by_name[name]
        if id is not None and name is None:
            if id in self.kis:
                return self.kis[id]
        return None

    def start_handle_loop(self, loops=None):
        did_loops = 0
        while loops is not None and did_loops < loops or loops is None:
            status, maybe_ke_request = self.long_poll()
            did_loops += 1
            if status == "repoll":
                continue
            elif status == "reregister":
                self.reregister()
            elif status == "exit":
                log.warn("KE returned status code 410, meaning I have to exit")
                break
            elif status == "handle":
                ke_request = maybe_ke_request

                # Extract relevant data items from the KE request
                ki_id = ke_request["knowledgeInteractionId"]
                handle_request_id = ke_request["handleRequestId"]
                bindings = ke_request["bindingSet"]
                requesting_kb_id = ke_request["requestingKnowledgeBaseId"]

                ki = self.kis[ki_id]
                if isinstance(ki, knowledge_interaction.AnswerKnowledgeInteraction):
                    result_bindings = ki.answer(bindings, requesting_kb_id)
                elif isinstance(ki, knowledge_interaction.ReactKnowledgeInteraction):
                    result_bindings = ki.react(bindings, requesting_kb_id)

                self.post_handle_response(ki, handle_request_id, result_bindings)

    def long_poll(self):
        log.info("Waiting for response to long poll...")
        response = requests.get(
            f"{self.ke_url}/sc/handle", headers={"Knowledge-Base-Id": self.id}
        )
        if response.status_code == 202:
            log.info("Received 202.")
            return "repoll", None
        elif response.status_code == 500:
            log.error(response.text)
            log.error("TKE had an internal server error. Reinitiating long poll.")
            return "repoll", None
        elif response.status_code == 410:
            return "exit", None
        elif response.status_code == 200:
            return "handle", response.json()
        elif response.status_code == 404:
            log.warn("Our Knowledge Base has been unregistered!")
            raise KnowledgeBaseUnregistered()
        else:
            log.warn(f"long_poll received unexpected status {response.status_code}")
            log.warn(response.text)
            log.warn("repolling anyway..")
            return "repoll", None

    def post_handle_response(self, ki, handle_id, bindings):
        response = requests.post(
            f"{self.ke_url}/sc/handle",
            json={
                "handleRequestId": handle_id,
                "bindingSet": bindings,
            },
            headers={
                "Knowledge-Base-Id": self.id,
                "Knowledge-Interaction-Id": ki.id,
            },
        )

        if not response.ok:
            raise UnexpectedHttpResponseError(response)



import logging as log
import mysql.connector
import time
from datetime import datetime

from knowledge_mapper.auth.base_auth import BaseAuth

MAX_CONNECTION_ATTEMPTS = 10
WAIT_BEFORE_RETRY = 2

class SqlAuth(BaseAuth):
    def __init__(self, conf: dict):
        self.host = conf['sql']['host']
        self.port = conf['sql']['port']
        self.database = conf['sql']['database']
        self.user = conf['sql']['user']
        self.password = conf['sql']['password']
        self.connect()

    def connect(self, max_attempts=MAX_CONNECTION_ATTEMPTS, wait_between_attempts=WAIT_BEFORE_RETRY):
        success = False
        attempts = 0
        while not success:
            try:
                log.info(f'Attempting SQL connection to {self.database} in {self.host}:{self.port} for SQL auth...')
                self.conn = mysql.connector.connect(
                    user=self.user,
                    password=self.password,
                    host=self.host,
                    port=self.port,
                    database=self.database
                )
                self.ping()
                success = True
            except Exception as e:
                attempts += 1
                if attempts < max_attempts:
                    log.warning(f'SQL connection attempt to {self.database} in {self.host}:{self.port} failed. Retrying in {wait_between_attempts} s.')
                    time.sleep(wait_between_attempts)
                else:
                    log.error(f'Failed to establish SQL connection.')
                    raise e
        log.info('Successfully made SQL connection for SQL auth.')
    
    def ping(self):
        self.conn.ping()

    def has_permission(self, kb_id, ki):
        cursor = self.conn.cursor(named_tuple=True)
        cursor.execute('SELECT COUNT(*) > 0 AS permission FROM policies WHERE (knowledge_interaction_id=%s OR knowledge_interaction_id IS NULL) AND (knowledge_base_id=%s OR knowledge_base_id IS NULL);', (ki['id'], kb_id))
        row = cursor.fetchone()
        if row is None:
            raise Exception('Expected a row in the result set.')
        cursor.close()
        self.conn.commit()
        if row.permission:
            log.debug(f'KB {kb_id} has access to KI {ki["id"]}')
        else:
            log.debug(f'KB {kb_id} does not have access to KI {ki["id"]}')

        cursor = self.conn.cursor()
        cursor.execute('INSERT INTO access_log (knowledge_interaction_id, knowledge_base_id, access_datetime, success) VALUES (%s, %s, %s, %s);', (ki['id'], kb_id, datetime.now(), row.permission))
        self.conn.commit()

        return row.permission
    
    def add_knowledge_interaction(self, ki):
        cursor = self.conn.cursor()
        cursor.execute('INSERT INTO knowledge_interactions (id, name) VALUES (%s, %s) ON DUPLICATE KEY UPDATE name=%s;', (ki['id'], ki['name'], ki['name']))
        self.conn.commit()



class BaseAuth:
    def has_permission(self, kb_id, ki):
        raise NotImplementedError("Please implement this abstract method.")
    
    def add_knowledge_interaction(self, ki):
        """Called when a Knowledge Interaction is added to this Knowledge Mapper"""
        pass



import requests
from requests.auth import HTTPBasicAuth
import os
import logging as log
from urllib.parse import quote

from .data_source import DataSource


class SparqlSource(DataSource):
    def __init__(self, endpoint: str, env_username, env_password):
        self.endpoint = endpoint
        self.auth = False
        # env_username and env_password MUST be the names of the environment variables
        # that contain the username and password to get access to the endpoint
        # if they are present, the self.auth flag will go up
        if (env_username != None and env_username in os.environ) and (
            env_password != None and env_password in os.environ
        ):
            self.auth = True
            self.username = os.environ[env_username]
            self.password = os.environ[env_password]

    def test(self):
        log.info("Testing SPARQL endpoint.")
        self.do_sparql_select("SELECT * WHERE { ?s ?p ?o . } LIMIT 0")
        log.info("Succes!")

    def handle(self, ki, binding_set, requesting_kb):
        if ki["type"] == "answer":
            return self.handle_answer(ki, binding_set, requesting_kb)
        elif ki["type"] == "react":
            return self.handle_react(ki, binding_set, requesting_kb)

    def handle_answer(self, ki, binding_set, requesting_kb):
        # Generate the SPARQL query based on the incoming bindings and the knowledge interaction's graph pattern.
        query = generate_sparql_select(ki, binding_set)
        # Fire the SPARQL query.
        result = self.do_sparql_select(query)
        # Restructure the bindings into TKE bindings and return it
        return restructure_bindings(result)

    def handle_react(self, ki, binding_set, requesting_kb):
        # Generate the SPARQL query based on the incoming bindings and the knowledge interaction's graph pattern.
        query = generate_sparql_insert(ki, binding_set)
        # Fire the SPARQL query.
        self.do_sparql_insert(query)
        # Restructure the bindings into TKE bindings and return it
        return []

    def do_sparql_select(self, query):
        args = {
            "headers": {
                "Accept": "application/sparql-results+json",
                "Content-Type": "application/sparql-query",
            }
        }

        if self.auth:
            args["auth"] = HTTPBasicAuth(self.username, self.password)

        response = requests.post(f"{self.endpoint}/query", data=query, **args)

        if response.status_code == 401:
            raise UnauthorizedError(
                "Provide BasicAuth with system environment variables."
            )
        elif not response.ok:
            raise RuntimeError(
                "Invalid response from SPARQL endpoint.  (status: {}, body: {})".format(
                    response.status_code, response.text
                )
            )

        return response.json()

    def do_sparql_insert(self, query):
        args = {
            "headers": {
                "Accept": "application/json",
                "Content-Type": "application/sparql-update",
            }
        }

        if self.auth:
            args["auth"] = HTTPBasicAuth(self.username, self.password)

        response = requests.post(f"{self.endpoint}", data=query, **args)

        if response.status_code == 401:
            raise UnauthorizedError(
                "Provide BasicAuth with system environment variables with the names used in the config file."
            )
        elif not response.ok:
            raise RuntimeError(
                "Invalid response from SPARQL endpoint.  (status: {}, body: {})".format(
                    response.status_code, response.text
                )
            )

        return


def generate_sparql_select(ki, incoming_bindings):
    # For all partial bindings that are actually partial, we set the
    # variables that ARE in the graph pattern, but NOT in the binding to
    # UNDEF, so that they match anything.
    for var in ki["vars"]:
        for incoming_binding in incoming_bindings:
            if var not in incoming_binding:
                incoming_binding[var] = "UNDEF"

    if "prefixes" in ki:
        prefixes = ki["prefixes"]
    else:
        prefixes = dict()

    return """
        {prefixes_clause}
        SELECT {{variables}} WHERE {{{{
            {triple_pattern}
            {values_clause}
        }}}}
    """.format(
        prefixes_clause="\n\t".join(
            f"PREFIX {prefix}: <{ki['prefixes'][prefix]}>" for prefix in prefixes.keys()
        ),
        triple_pattern=ki["pattern"],
        values_clause="""
                VALUES ({{variables}}) {{{{
                    {bindings}
                }}}}
            """.format(
            bindings="\n".join(
                [
                    f'({" ".join([binding[var] for var in ki["vars"]])})'
                    for binding in incoming_bindings
                ]
            )
        )
        if incoming_bindings
        else "",
    ).format(
        variables=" ".join([f"?{var}" for var in ki["vars"]]),
    )


def generate_sparql_insert(ki, bindings):
    variables = ki["vars"]
    if "prefixes" in ki:
        prefixes = ki["prefixes"]
    else:
        prefixes = dict()
    pref = "\n\t".join(
        f"PREFIX {prefix}: <{ki['prefixes'][prefix]}>" for prefix in prefixes.keys()
    )

    return f"""
        {pref}
        INSERT {{
            {ki['argument_pattern']}
        }} WHERE {{ VALUES ({' '.join([f'?{variable}' for variable in variables])}) {{
            {os.linesep.join(f'({" ".join([str(binding[variable]) for variable in variables])})' for binding in bindings)}
        }} }}
    """


def restructure_bindings(sparql_results):
    restructured_binding_set = []
    for binding in sparql_results["results"]["bindings"]:
        restructured_binding = dict()
        for key, value in binding.items():
            if value["type"] == "uri":
                restructured_value = f'<{value["value"]}>'
            elif binding[key]["type"] == "literal":
                if "datatype" in value:
                    restructured_value = f'"{value["value"]}"^^<{value["datatype"]}>'
                else:
                    restructured_value = f'"{value["value"]}"'
            restructured_binding[key] = restructured_value

        restructured_binding_set.append(restructured_binding)

    return restructured_binding_set


class UnauthorizedError(RuntimeError):
    pass



from setuptools import setup, find_packages

setup(
    name="knowledge_mapper",
    version="0.0.19",
    packages=find_packages(),
    install_requires=["requests", "mysql-connector-python", "json5"],
    entry_points={
        "console_scripts": [
            "knowledge_mapper=knowledge_mapper.__main__:main",
        ]
    },
)
